--[[
Copyright (C) GtX (Andy), 2017

Interface: 1.5.1.0 b1580

Author: GtX (LS-Modcompany)
Date: 01.12.2017

Support only at: http://ls-modcompany.com

History:
CLOSED BETA-1 @ 09.12.2017

Version: 1.0.0.0 - Release Edition @ 07.12.2017
Version: 1.1.0.0 - Update @ 16.02.2018 	- Added new GUI and option to adjust the detection strength of the display. Useful if the display is detecting more displays then you want.
											Hub Display can now support unlimited number of fillTypes. Scrolling is now possible using the `mouse wheel` or `PageUp/PageDown Keys`.
											('mouse wheel' will only work if no handTools are in your kit, otherwise it is disabled.)

Important:
No changes are to be made to this script without permission from GtX @ http://ls-modcompany.com

** PLACEABLE ONLY! **

Notes:
The script can target multiple silo triggers at the same time. Placing two Hay Lofts side by side will allow you to see the combined total for example.
--]]

SiloDisplay = {}
SiloDisplay.ModDir = g_currentModDirectory
SiloDisplay_mt = Class(SiloDisplay, Placeable)
source(Utils.getFilename("scripts/SiloDisplayGui.lua", SiloDisplay.ModDir))

InitObjectClass(SiloDisplay, "SiloDisplay")

local function get_i18n(name)
	return g_i18n:hasText(name) and g_i18n:getText(name) or name
end

function SiloDisplay:new(isServer, isClient, customMt)
    local mt = customMt
    if mt == nil then
        mt = SiloDisplay_mt
    end

    local self = Placeable:new(isServer, isClient, mt)

    registerObjectClassName(self, "SiloDisplay")

    return self
end

function SiloDisplay:load(xmlFilename, x,y,z, rx,ry,rz, initRandom)

	self.nodeId = xmlFilename
	self.seekRange = 50

	if not SiloDisplay:superClass().load(self, xmlFilename, x,y,z, rx,ry,rz, initRandom) then
		return false
	end

	return true

end

function SiloDisplay:finalizePlacement(x, y, z, rx, ry, rz, initRandom)
	SiloDisplay:superClass().finalizePlacement(self)

	self.siloFilledLevel = {}
	self.siloCurrentLevel = {}
	self.systemConnected = false
	self.playerInTrigger = false
	self.canInteract = false
	self.vehiclesInRange = {}
	self.siloInRange = nil
	self.firstTimeRun = true
	self.placedSeek = true
	self.fillPageState = 0

	self.rangeTable = {}
	local setRange = 0
	local rangeLimit = 100
	local text = get_i18n("warning_siloRangeSelect")
	for r=1, rangeLimit do
		setRange = setRange + 1
		local rangeNumber = string.format("%s %s", text, tostring(setRange).."%")
		table.insert(self.rangeTable, rangeNumber)
	end

	self.hubOnly = Utils.getNoNil(getUserAttribute(self.nodeId, "hubOnly"), false) -- Hub Only does not use displays. Rather it uses a GUI Display.
	self.showFreeCapacity = Utils.getNoNil(getUserAttribute(self.nodeId, "showFreeCapacity"), false) -- If 'true' this will display the remaining silo capacity rather than the total.

	if self.hubOnly then
		self.hideCircle = false

		local uiScale = g_gameSettings:getValue("uiScale")
		local width, height = getNormalizedScreenValues(500*uiScale, 600*uiScale)
		local loadingWidth, loadingHeight = getNormalizedScreenValues(500*uiScale, 500*uiScale)
		self.hubHudBack = Overlay:new("hubHudBack", g_baseUIFilename, 0.5,0.45, width, height)
		self.hubHudBack:setAlignment(Overlay.ALIGN_VERTICAL_MIDDLE, Overlay.ALIGN_HORIZONTAL_CENTER)
		self.hubHudBack:setUVs(g_colorBgUVs)
		self.hubHudBack:setColor(unpack(g_colorBg))

		local _, headerHeight = getNormalizedScreenValues(0, 30*uiScale)
		self.hubHudHead = Overlay:new("hubHudHead", g_baseUIFilename, 0.5,0.45+height / 2, width, headerHeight)
		self.hubHudHead:setAlignment(Overlay.ALIGN_VERTICAL_TOP, Overlay.ALIGN_HORIZONTAL_CENTER)
		self.hubHudHead:setUVs(g_colorBgUVs)
		self.hubHudHead:setColor(0.0075, 0.0075, 0.0075, 1)

		self.hubHudSpace = Overlay:new("hubHudSpace", g_baseUIFilename, 0.5,0.45, width, 2 / g_screenHeight)
		self.hubHudSpace:setAlignment(Overlay.ALIGN_VERTICAL_BOTTOM, Overlay.ALIGN_HORIZONTAL_CENTER)
		self.hubHudSpace:setUVs(g_colorBgUVs)
		self.hubHudSpace:setColor(0.0075, 0.0075, 0.0075, 1)

		local hubHudFile = getUserAttribute(self.nodeId,"updateGuiScreenFile")
		if hubHudFile and hubHudFile ~= 0 then
			local hubHudLoadingFile = Utils.getFilename(hubHudFile, SiloDisplay.ModDir)
			self.hubHudLoading = Overlay:new("hubHudLoading", hubHudLoadingFile, 0.5,0.5, loadingWidth, loadingHeight)
			self.hubHudLoading:setAlignment(Overlay.ALIGN_VERTICAL_MIDDLE, Overlay.ALIGN_HORIZONTAL_CENTER)
		end
	end

	local DisplaysIndex = getUserAttribute(self.nodeId, "displayGroup")
	if DisplaysIndex ~= nil then
		self.displays = {}
		local DisplaysId = Utils.indexToObject(self.nodeId, DisplaysIndex)

		if DisplaysId and DisplaysId ~= 0 then

			local mainDisplayIndex = getUserAttribute(DisplaysId, "mainDisplays")
			if mainDisplayIndex ~= nil then
				self.displays.mainDisplay = Utils.indexToObject(DisplaysId, mainDisplayIndex)

				-- Only done on load as no update needed after game is loaded as language does not change.
				local useLanguage = Utils.getNoNil(getUserAttribute(self.displays.mainDisplay, "useLanguageNodes"), false)
				if useLanguage then
					self.mainScreenDisplay = {}
					local langId = getNumOfChildren(self.displays.mainDisplay)
					for lt=1,langId do
						local language = getName(getChildAt(self.displays.mainDisplay,lt-1))
						self.mainScreenDisplay[language] = {}
						self.mainScreenDisplay[language].node = getChildAt(self.displays.mainDisplay,lt-1)
						setVisibility(self.mainScreenDisplay[language].node, false)
					end

					if self.mainScreenDisplay ~= nil then
						local englishBase = "en"
						if self.mainScreenDisplay[englishBase] ~= nil then
							if self.mainScreenDisplay[g_languageShort] ~= nil then
								setVisibility(self.mainScreenDisplay[g_languageShort].node, true)
							else
								setVisibility(self.mainScreenDisplay[englishBase].node, true)
							end
						else
							setVisibility(self.displays.mainDisplay, false)
							print("ERROR: [ SiloDisplay.lua ] English(en) node must be set as a backup for all other languages not set as transform groups under the Attribute 'mainDisplays' if using Attribute 'useLanguageNodes' -  All items hidden!!")
						end
					end
				end

			end

			local NumConnectedIndex = getUserAttribute(DisplaysId, "connectedCounter")
			if NumConnectedIndex ~= nil then
				local NumConnected = Utils.indexToObject(DisplaysId, NumConnectedIndex)
				local connected = 0
				self.displays.connected = {}
				self.displays.connected.node = NumConnected
				Utils.setNumberShaderByValue(self.displays.connected.node, math.floor(connected), 0, true)
			end

			local SearchDisplayIndex = getUserAttribute(DisplaysId, "searchDisplay")
			if SearchDisplayIndex ~= nil then
				local SearchDisplay = Utils.indexToObject(DisplaysId, SearchDisplayIndex)
				self.displays.search = SearchDisplay
				setVisibility(self.displays.search, false)
			end

			if not self.hubOnly then
				local levelDisplays = Utils.indexToObject(DisplaysId, getUserAttribute(DisplaysId, "levelDisplays"))
				local levelCount = getNumOfChildren(levelDisplays)
				self.displays.levels = {}

				for i=1, levelCount do
					local fillTypeName = getName(getChildAt(levelDisplays, i-1))
					local fillTypInt = FillUtil.fillTypeNameToInt[fillTypeName]

					if fillTypInt ~= nil and fillTypInt ~= FillUtil.FILLTYPE_UNKNOWN then
						local fillTypeLevel = 0
						self.siloFilledLevel[fillTypInt] = -100
						self.displays.levels[fillTypInt] = {}
						self.displays.levels[fillTypInt].node = getChildAt(levelDisplays, i-1)

						Utils.setNumberShaderByValue(self.displays.levels[fillTypInt].node, math.floor(fillTypeLevel), 0, true)
						setVisibility(self.displays.levels[fillTypInt].node, false)
					end
				end

				local capacityDisplayIndex = getUserAttribute(DisplaysId, "capacityDisplays")
				if capacityDisplayIndex ~= nil then
					local capacityDisplay = Utils.indexToObject(DisplaysId, capacityDisplayIndex)
					local capacityCount = getNumOfChildren(capacityDisplay)
					self.displays.capacity = {}

					for i=1, capacityCount do
						local fillTypeName = getName(getChildAt(capacityDisplay, i-1))
						local fillTypInt = FillUtil.fillTypeNameToInt[fillTypeName]

						if fillTypInt ~= nil and fillTypInt ~= FillUtil.FILLTYPE_UNKNOWN then
							local siloCapacity = 0
							self.displays.capacity[fillTypInt] = {}
							self.displays.capacity[fillTypInt].node = getChildAt(capacityDisplay, i-1)

							Utils.setNumberShaderByValue(self.displays.capacity[fillTypInt].node, math.floor(siloCapacity), 0, true)
							setVisibility(self.displays.capacity[fillTypInt].node, false)
						end
					end
				end

				local labelDisplays = getUserAttribute(DisplaysId, "fillTypeLabels")
				if labelDisplays ~= nil then
					self.labels = {}
					local labelNames = Utils.indexToObject(DisplaysId, labelDisplays)
					local typeCount = getNumOfChildren(labelNames)
					for i=1, typeCount do
						local fillTypeName = getName(getChildAt(labelNames, i-1))
						local fillTypInt = FillUtil.fillTypeNameToInt[fillTypeName]
						if fillTypInt ~= nil and fillTypInt ~= FillUtil.FILLTYPE_UNKNOWN then
							local fillNode = getChildAt(labelNames, i-1)
							self.labels[fillTypInt] = {}
							self.labels[fillTypInt].node = fillNode
							setVisibility(self.labels[fillTypInt].node, false)

							-- Only done on load as no update needed after game is loaded as language does not change.
							local useLanguage = Utils.getNoNil(getUserAttribute(labelNames, "useLanguageNodes"), false)
							if useLanguage then
								local languageTypeId = fillNode
								local languageType = {}
								local langId = getNumOfChildren(languageTypeId)
								for lt=1,langId do
									local language = getName(getChildAt(languageTypeId,lt-1))
									languageType[language] = {}
									languageType[language].node = getChildAt(languageTypeId,lt-1)
									setVisibility(languageType[language].node, false)
								end

								if languageType ~= nil then
									local englishBase = "en"
									if languageType[englishBase] ~= nil then
										if languageType[g_languageShort] ~= nil then
											setVisibility(languageType[g_languageShort].node, true)
										else
											setVisibility(languageType[englishBase].node, true)
										end
									else
										setVisibility(languageTypeId, false)
										print("ERROR: [ SiloDisplay.lua ] English(en) node must be set as a backup for all other languages not set as transform groups under the Attribute 'fillTypeLabels' if using Attribute 'useLanguageNodes' -  All items hidden!!")
									end
								end
							end
						end
					end
				end
			end
		end
	end

	local PlayerIndex = getUserAttribute(self.nodeId, "playerTriggerIndex")
	if PlayerIndex ~= nil then
		if self.hubOnly then
			self.showNameHud = false
		else
			self.showNameHud = Utils.getNoNil(getUserAttribute(self.nodeId, "showStorageNameInHud"), true)
		end

		local PlayerTriggerId = Utils.indexToObject(self.nodeId, PlayerIndex)

		if PlayerTriggerId and PlayerTriggerId ~= 0 then
			self.playerTrigger = PlayerTriggerId
			addTrigger(self.playerTrigger,"playerTriggerCallback", self)
		end
	end

	if not self.hubOnly then
		local UpdateIndex = getUserAttribute(self.nodeId, "updateTriggerIndex")
		if UpdateIndex ~= nil then
			local UpdateTriggerId = Utils.indexToObject(self.nodeId, UpdateIndex)

			if UpdateTriggerId and UpdateTriggerId ~= 0 then
				self.updateTrigger = UpdateTriggerId
				addTrigger(self.updateTrigger,"updateTriggerCallback", self)
			end
		end
	end

    return true

end

function SiloDisplay:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
    if not SiloDisplay:superClass().loadFromAttributesAndNodes(self, xmlFile, key, resetVehicles) then
        return false
    end

	local seekRange = getXMLInt(xmlFile, key.."#setSeekRange")
    if seekRange ~= nil and seekRange >= 0 and seekRange <= 100 then
        self.seekRange = seekRange
		self.placedSeek = false
	else
		self.seekRange = 50
		self.placedSeek = true
    end

    return true
end

function SiloDisplay:getSaveAttributesAndNodes(nodeIdent)
    local attributes, nodes = SiloDisplay:superClass().getSaveAttributesAndNodes(self, nodeIdent)

	attributes = attributes .. ' setSeekRange="'..self.seekRange..'"'

	local n = 0
	if self.siloInRange ~= nil then
		for _, siloTrigger in pairs(self.siloInRange) do
			for siloSource,_ in pairs(siloTrigger.siloSources) do
				if n > 0 then
					nodes = nodes.."\n"
				end
				local rootNode = tostring(siloSource.rootNode)
				local saveId = siloSource.saveId

				nodes = nodes..nodeIdent..'<LinkedStorage rootNode="'..rootNode..'" saveId="'..saveId..'" />'

				n = n + 1
			end
		end
	else
		local linkedStorageId = "N/A"
		local linkedStorageSaveId = "No_Storage_In_Range"
		nodes = nodes..nodeIdent..'<LinkedStorage rootNode="'..linkedStorageId..'" saveId="'..linkedStorageSaveId..'" />'
	end

    return attributes, nodes
end

function SiloDisplay:delete()
    unregisterObjectClassName(self)
    SiloDisplay:superClass().delete(self)

	if self.playerTrigger then
		removeTrigger(self.playerTrigger)
	end

	if self.updateTrigger then
		removeTrigger(self.updateTrigger)
	end

	if self.hubOnly then
		self.hubHudBack:delete()
		self.hubHudHead:delete()
		self.hubHudSpace:delete()
		self.hubHudLoading:delete()
	end
end

function SiloDisplay:readStream(streamId, connection)
    SiloDisplay:superClass().readStream(self, streamId, connection)
    if connection:getIsServer() then
		local placedSeek = streamReadBool(streamId)
		if not placedSeek then
			self.placedSeek = placedSeek
			self.seekRange = streamReadInt16(streamId)
		end
    end
end

function SiloDisplay:writeStream(streamId, connection)
    SiloDisplay:superClass().writeStream(self, streamId, connection)
    if not connection:getIsServer() then
		if not self.placedSeek then
			streamWriteBool(streamId, self.placedSeek)
			streamWriteInt16(streamId, self.seekRange)
		end
    end
end

function SiloDisplay:update(dt)
	if self:getCanUpdate() or self.firstTimeRun then

		-- Only on load and silo search.
		if self.firstTimeRun then
			self.firstTimeRun = false
			if self.placedSeek then
				local seekRange = self:getTargetInRange(true)
				self.seekRange = math.floor(seekRange)+1
				self.placedSeek = false
			end

			self.siloInRange = self:getTargetInRange(false)

			if self.siloInRange ~= nil then
				if self.displays ~= nil then
					if self.displays.connected ~= nil then
						local connected = #self.siloInRange
						Utils.setNumberShaderByValue(self.displays.connected.node, math.floor(connected), 0, true)
					end

					if self.displays.mainDisplay ~= nil then
						if not getVisibility(self.displays.mainDisplay) then
							setVisibility(self.displays.mainDisplay, true)
						end
					end

					if self.displays.search ~= nil then
						if getVisibility(self.displays.search) then
							setVisibility(self.displays.search, false)
						end
					end

					if not self.hubOnly then
						if self.displays.levels ~= nil then
							for typ,_ in pairs(self.displays.levels) do
								self:updateLevelDisplays(0, typ, true)
							end
						end
					end
				end

				self.systemConnected = true

			else

				if self.displays ~= nil then
					if self.displays.mainDisplay ~= nil then
						if getVisibility(self.displays.mainDisplay) then
							setVisibility(self.displays.mainDisplay, false)
						end
					end

					if self.displays.search ~= nil then
						if not getVisibility(self.displays.search) then
							setVisibility(self.displays.search, true)
						end
					end

					if self.displays.connected ~= nil then
						local connected = 0
						Utils.setNumberShaderByValue(self.displays.connected.node, math.floor(connected), 0, true)
					end
				end

				self.systemConnected = false

			end
		end

		-- Only on load, silo search and when player is in range.
		if self.siloInRange ~= nil then
			self.siloCurrentLevel = {}

			for _, siloTrigger in pairs(self.siloInRange) do
				for siloSource,_ in pairs(siloTrigger.siloSources) do
					for fillType, fillLevel in pairs(siloSource.fillLevels) do
						if self.hubOnly then
							self.siloCurrentLevel[fillType] = Utils.getNoNil(self.siloCurrentLevel[fillType], 0) + fillLevel
						else
							if self.siloFilledLevel[fillType] ~= nil then
								self.siloCurrentLevel[fillType] = Utils.getNoNil(self.siloCurrentLevel[fillType], 0) + fillLevel
							end
						end
					end
				end
			end

			if not self.hubOnly then
				for fillType, fillLevel in pairs (self.siloCurrentLevel) do
					if fillType ~= nil and fillLevel ~= nil then
						if self.siloFilledLevel[fillType] ~= fillLevel then
							self.siloFilledLevel[fillType] = fillLevel
							if self.displays ~= nil then
								self:updateLevelDisplays(fillLevel, fillType, false)
							end
						end
					end
				end
			end
		end
	end

	-- Is player in trigger?
	if self.canInteract then
		g_currentMission:addHelpButtonText(get_i18n("input_SiloDisplay_Reset"), InputBinding.SiloDisplay_Reset)
		if InputBinding.hasEvent(InputBinding.SiloDisplay_Reset) then
			if g_siloDisplayGui == nil then
				g_siloDisplayGui = SiloDisplayGui:new()
				g_gui:loadGui(SiloDisplay.ModDir .. "scripts/SiloDisplayGui.xml", "SiloDisplayGui", g_siloDisplayGui)
			end
			self:openResetMenu()
		end

		if self.systemConnected then
			if self.showNameHud then
				local nameToDisplay, inRange = self:getTargetNameData()

				if inRange ~= 0 then
					if inRange == 1 then
						g_currentMission:addExtraPrintText("DISPLAY:  "..get_i18n(nameToDisplay))
					else
						g_currentMission:addExtraPrintText("DISPLAY:  "..get_i18n(nameToDisplay).." x "..inRange)
					end
				end
			end
		else
			g_currentMission:addExtraPrintText(g_i18n:getText("warning_storageDeleted"))
		end

		if self.hubOnly then
			if not g_gui:getIsGuiVisible() then
				self:drawHubHud()

				-- Allow scrolling of Hub Display. This allows unlimited Fruit Types ;-)
				local scrollNeeded = table.getn(self.levelToShow) > 17
				if scrollNeeded then
					local scrollUp = InputBinding.hasEvent(InputBinding.SiloDisplay_Up)
					local scrollDown = InputBinding.hasEvent(InputBinding.SiloDisplay_Down)
					local haveHandTools = table.getn(g_currentMission.player.tools) > 0
					if not haveHandTools then
						g_currentMission:addHelpButtonText(get_i18n("input_MENU_AXIS_UP_DOWN"), InputBinding.CAMERA_ZOOM_IN,InputBinding.CAMERA_ZOOM_OUT)
						scrollUp = InputBinding.hasEvent(InputBinding.CAMERA_ZOOM_IN) or InputBinding.hasEvent(InputBinding.SiloDisplay_Up)
						scrollDown = InputBinding.hasEvent(InputBinding.CAMERA_ZOOM_OUT) or InputBinding.hasEvent(InputBinding.SiloDisplay_Down)
					end
					g_currentMission:addHelpButtonText(get_i18n("input_MENU_AXIS_UP_DOWN"), InputBinding.SiloDisplay_Up,InputBinding.SiloDisplay_Down)
					if scrollDown then
						if (self.fillPageState + 17) < table.getn(self.levelToShow) then
							self.fillPageState = self.fillPageState + 1
						end
					elseif scrollUp then
						self.fillPageState = self.fillPageState - 1
						if self.fillPageState < 0 then
							self.fillPageState = 0
						end
					end
				end
			end
		end

		-- Remove 'player pointer' when UI is open.
		if self.hideCircle then
			if g_currentMission.controlPlayer then
				if g_currentMission.player ~= nil then
					g_currentMission.player.pickedUpObjectOverlay:setIsVisible(false)
				end
			end
		end

		if g_currentMission.controlledVehicle ~= nil then
			self.canInteract = false
		end

	else
		-- Reset 'player pointer'. Added support for GameExtension mod so we do not break settings.
		-- Only mod reported that changes the pointer ;-)
		if self.hubOnly then
			if self.hideCircle then
				if g_currentMission.controlPlayer then
					if g_currentMission.player ~= nil then
						if g_currentMission.showHudEnv then
							-- Return the same settings as user setup in Game Extension if the mod is installed.
							if g_currentMission.GameExtension ~= nil then
								if (g_currentMission.GameExtension.getSetting("PLAYER", "CROSSHAIR")) then
									g_currentMission.player.pickedUpObjectOverlay:setIsVisible(true)
								else
									g_currentMission.player.pickedUpObjectOverlay:setIsVisible(false)
								end
							else
								g_currentMission.player.pickedUpObjectOverlay:setIsVisible(true)
							end
						end
					end
				end

				self.hideCircle = false

			end
		end
	end

end

function SiloDisplay:openResetMenu()
	local gui = g_gui:showDialog("SiloDisplayGui")
	if gui ~= nil then
		gui.target:setTitle("TECH FARM")
		gui.target:setDisplayRange(self.rangeTable, self.seekRange, self.nodeId)
		gui.target:loadCallback(self.callbackSorting, self)
	end
end

function SiloDisplay:callbackSorting(newRange)
	if g_currentMission:getIsServer() then
		self.seekRange = newRange
		self.firstTimeRun = true
	else
		g_client:getServerConnection():sendEvent(SiloDisplayResetEvent:new(self, true, newRange))
	end
end

function SiloDisplay:drawHubHud()
	if g_currentMission.inGameMessage:getIsVisible() then
		return
	end

	self.hideCircle = true

	local nameToDisplay, inRange = self:getTargetNameData()
	local uiScale = g_gameSettings:getValue("uiScale")
	local _, textSize = getNormalizedScreenValues(0, 12*uiScale)
	local textOffsetX, textOffsetY = getNormalizedScreenValues(8*uiScale, 19*uiScale)
	local width, height = getNormalizedScreenValues(500*uiScale, 600*uiScale)
	local _, headerHeight = getNormalizedScreenValues(0, 30*uiScale)

	self.hubHudBack:render()
	self.hubHudHead:render()
	setTextColor(1,1,1,1)
	setTextAlignment(RenderText.ALIGN_CENTER)
	setTextBold(true)

	if inRange ~= 0 then
		if inRange == 1 then
			renderText(self.hubHudHead.x, self.hubHudHead.y-textOffsetY, textSize, get_i18n(nameToDisplay))
		else
			renderText(self.hubHudHead.x, self.hubHudHead.y-textOffsetY, textSize, get_i18n(nameToDisplay).." x "..inRange)
		end

		self.levelToShow = {}
		local text = nil
		for fillType, fillLevel in pairs (self.siloCurrentLevel) do
			local capacity = self:getCapacity(fillType)
			local fillTypeDesc = FillUtil.fillTypeIndexToDesc[fillType]
			local capacityAdjust = fillLevel

			if self.showFreeCapacity then
				capacityAdjust = 0
			end

			text = {}
			text.name = fillTypeDesc.nameI18N
			text.fillType = string.format("%s : %s  /  %s %s", g_i18n:getText("info_fillLevel"), math.floor(Utils.getNoNil(fillLevel,0)), math.floor(Utils.getNoNil(capacity+capacityAdjust,0)), g_i18n:getText("unit_literShort"))
			text.capacity = capacity

			table.insert(self.levelToShow, text)
		end

		setTextBold(false)
		setTextAlignment(RenderText.ALIGN_LEFT)
		local numberFills = table.getn(self.levelToShow)

		if numberFills > 0 then
			self.hubHudSpace:setPosition(nil, self.hubHudHead.y-headerHeight - headerHeight)
			self.hubHudSpace:render()

			for i=1, math.min(numberFills, 17) do
				local v = self.levelToShow[i + self.fillPageState]

				renderText(self.hubHudHead.x-width/2.2, self.hubHudHead.y-textOffsetY - (i+1)*headerHeight, textSize, v.name)
				renderText(self.hubHudHead.x-width/7, self.hubHudHead.y-textOffsetY - (i+1)*headerHeight, textSize, v.fillType)

				self.hubHudSpace:setPosition(nil, self.hubHudHead.y-headerHeight - (i+1)*headerHeight)
				self.hubHudSpace:render()
			end

			setTextBold(true)
			setTextColor(0.2122, 0.5271, 0.0307, 1)
			setTextAlignment(RenderText.ALIGN_CENTER)
			if self.fillPageState > 0 then
				renderText(self.hubHudHead.x, self.hubHudHead.y-textOffsetY - 1*headerHeight, textSize, get_i18n("hub_fillTypeUp"))
			end
			if (self.fillPageState + 17) < numberFills then
				renderText(self.hubHudHead.x, self.hubHudHead.y-textOffsetY - 19*headerHeight, textSize, get_i18n("hub_fillTypeDown"))
			end
			setTextBold(false)
		end

	else
		setTextBold(false)
		self.hubHudLoading:render()
		setTextAlignment(RenderText.ALIGN_CENTER)
		local setStatusText = "( 0 )"
		setTextColor(1,0,0,1)
		local _, textSize = getNormalizedScreenValues(0, 24*uiScale)
		renderText(0.5,0.3, textSize, setStatusText)
	end
end

function SiloDisplay:getTargetNameData()
	local name = ""
	local number = 0

	if self.siloInRange ~= nil then

		number = #self.siloInRange

		for _, siloTrigger in pairs(self.siloInRange) do
			if number > 1 then
				name = g_i18n:getText("helpIcon_title11")
			else
				name = siloTrigger.stationName
			end
		end
	end

	return name, number
end

function SiloDisplay:playerTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay)

	if (g_currentMission.controlPlayer and g_currentMission.player and otherId == g_currentMission.player.rootNode) then
		if (onEnter) then
			self.canInteract = true
		elseif (onLeave) then
			self.canInteract = false
		end
	end

end

function SiloDisplay:updateTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter or onLeave then
        if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
            if onEnter then
                self.playerInTrigger = true
            else
                self.playerInTrigger = false
            end
        else
            local vehicle = g_currentMission.nodeToVehicle[otherShapeId]
            if vehicle ~= nil then
                if onEnter then
                    if self.vehiclesInRange[vehicle] == nil then
                        self.vehiclesInRange[vehicle] = true
                    end
                else
                    if self.vehiclesInRange[vehicle] then
                        self.vehiclesInRange[vehicle] = nil
                    end
                end
            end
        end
    end
end

function SiloDisplay:getCanUpdate()
	if self.hubOnly then
		return self.canInteract
	else
		if (g_currentMission.controlPlayer and self.playerInTrigger) then
			return true
		end

		if not g_currentMission.controlPlayer then
			for vehicle in pairs(self.vehiclesInRange) do
				if vehicle:getIsActiveForInput(false) then
					return true
				end
			end
		end
	end

    return false
end

function SiloDisplay:getCanBePlacedAt(x, y, z, distance)
    local canBePlaced = SiloDisplay:superClass().getCanBePlacedAt(self, x, y, z, distance)

    if canBePlaced then
        local a,b,c = getTranslation(self.nodeId)
        setTranslation(self.nodeId, x,y,z)

        local inRange = self:getTargetInRange(false)

        if inRange == nil then
            canBePlaced = false
        end

        setTranslation(self.nodeId, a,b,c)

    end

    return canBePlaced
end

function SiloDisplay:getCapacity(fillType)
	local siloCapacity = 0
	for _, siloTrigger in pairs(self.siloInRange) do
		for siloSource,_ in pairs(siloTrigger.siloSources) do
			siloCapacity = Utils.getNoNil(siloCapacity, 0) + siloSource:getFreeCapacity(fillType)
		end
	end
	return siloCapacity
end

function SiloDisplay:updateLevelDisplays(updatedLevel, siloFillType, siloMoved)

	if siloFillType ~= nil and not siloMoved then
		if self.displays.levels[siloFillType].node ~= nil then
			Utils.setNumberShaderByValue(self.displays.levels[siloFillType].node, math.floor(updatedLevel), 0, true)
			if not getVisibility(self.displays.levels[siloFillType].node) then
				setVisibility(self.displays.levels[siloFillType].node, true)
			end
			if self.labels[siloFillType].node ~= nil then
				if not getVisibility(self.labels[siloFillType].node) then
					setVisibility(self.labels[siloFillType].node, true)
				end
			end
		end
	end

	if siloFillType ~= nil and siloMoved then
		if getVisibility(self.displays.levels[siloFillType].node) then
			setVisibility(self.displays.levels[siloFillType].node, false)
		end
		if self.labels[siloFillType].node ~= nil then
			if getVisibility(self.labels[siloFillType].node) then
				setVisibility(self.labels[siloFillType].node, false)
			end
		end
		self.siloFilledLevel[siloFillType] = -100
	end

	if self.displays.capacity ~= nil then
		local capacityAdjust = updatedLevel

		if self.showFreeCapacity then
			capacityAdjust = 0
		end

		self:updateCapacity(siloFillType, siloMoved, capacityAdjust)
	end
end

function SiloDisplay:updateCapacity(siloFillType, siloMoved, capacityAdjust)

	if siloFillType ~= nil and not siloMoved then

		Utils.setNumberShaderByValue(self.displays.capacity[siloFillType].node, math.floor(self:getCapacity(siloFillType) + capacityAdjust), 0, true)

		if not getVisibility(self.displays.capacity[siloFillType].node) then
			setVisibility(self.displays.capacity[siloFillType].node, true)
		end

	end

	if siloFillType ~= nil and siloMoved then
		if getVisibility(self.displays.capacity[siloFillType].node) then
			setVisibility(self.displays.capacity[siloFillType].node, false)
		end
	end

end

function SiloDisplay:getTargetInRange(placeSet)
	local siloTargets = nil

	local xNode,yNode,zNode = getWorldTranslation(self.nodeId)
    for target,_ in pairs(g_currentMission.siloTriggers) do
        local xSilo,ySilo,zSilo = getWorldTranslation(target.rootNode)
        local distance = Utils.vector3Length(xNode-xSilo, yNode-ySilo, zNode-zSilo)
        if distance < self.seekRange then
			if placeSet then
				siloTargets = distance
			else
				--Allow for multi triggers.
				if siloTargets == nil then
					siloTargets = {}
				else
					siloTargets = siloTargets
				end
				table.insert(siloTargets, target)
			end
        end
    end

    return siloTargets

end

registerPlaceableType("SiloDisplay", SiloDisplay)

-- Search / Reset Event
SiloDisplayResetEvent = {}
SiloDisplayResetEvent_mt = Class(SiloDisplayResetEvent, Event)
InitEventClass(SiloDisplayResetEvent, "SiloDisplayResetEvent")

function SiloDisplayResetEvent:emptyNew()
    local self = Event:new(SiloDisplayResetEvent_mt)
    self.className = "SiloDisplayResetEvent"
    return self
end

function SiloDisplayResetEvent:new(object, firstTimeRun, newRange)
    local self = SiloDisplayResetEvent:emptyNew()

    self.object = object
    self.firstTimeRun = firstTimeRun
	self.newRange = newRange

    return self
end

function SiloDisplayResetEvent:writeStream(streamId, connection)
    writeNetworkNodeObject(streamId, self.object)
    streamWriteBool(streamId, self.firstTimeRun)
	streamWriteInt16(streamId, self.newRange)
end

function SiloDisplayResetEvent:readStream(streamId, connection)
    self.object = readNetworkNodeObject(streamId)
    self.firstTimeRun = streamReadBool(streamId)
	self.newRange = streamReadInt16(streamId)

    self:run(connection)
end

function SiloDisplayResetEvent:run(connection)
    if not connection:getIsServer() then
		g_server:broadcastEvent(SiloDisplayResetEvent:new(self.object, self.firstTimeRun, self.newRange))
    end

    if self.object ~= nil then
		self.object.seekRange = self.newRange
        self.object.firstTimeRun = self.firstTimeRun
    end
end